;;; keys.el
;;;
;;; Define algumas teclas de atalho e funcoes

;; [Esc] [Esc] invoca goto-line
(define-key global-map "\e\e" 'goto-line)

;; Define a funcao dos Fx's
(global-set-key [f1] 'undo)
(global-set-key [f2] 'eshell)
(global-set-key [f3] 'liga-acentos)
(global-set-key [f4] 'call-last-kbd-macro)
(global-set-key [f5] 'bbdb-complete-name)

;; Page Up, Down, Home e End
(global-set-key [next] 'yic-next-buffer)
(global-set-key [prior] 'yic-prev-buffer)
(global-set-key [end] 'kill-buffer)
(global-set-key [home] 'other-window)

;; insere data formatada "'dia semana' 'mes' 'dia mes', 'ano' 'horario'"
(defun insert-date ()
  "Insere data."
  (interactive)
  (insert (format-time-string "%a %b %e, %Y @ %l:%M %p")))

;; funcao para ligar os acentos
(defun liga-acentos ()
  "Liga os acentos para linguagem portugues brasileiro."
  (interactive)
  (iso-accents-mode t)
  (iso-accents-customize "portuguese"))

;; Funcoes para troca de buffer
(defun yic-ignore (str)
  (or
   ;; Buffers que eu nao quero ver
   (string-match "\\*Buffer List\\*" str)
   (string-match "^TAGS" str)
   (string-match "^\\*Messages\\*$" str)
   (string-match "^\\*Completions\\*$" str)
   ;;   (string-match "^\\*scratch\\*$" str)
   (string-match "^ " str)
   
   (memq str		   
	 (mapcar 
	  (lambda (x) 
	    (buffer-name 
	     (window-buffer 
	      (frame-selected-window x))))
	  (visible-frame-list)))
   ))

(defun yic-next (ls)
  "Vai para o proximo buffer tirando os que estao em ignore."
  (let* ((ptr ls)
	 bf bn go
	 )
    (while (and ptr (null go))
      (setq bf (car ptr)  bn (buffer-name bf))
      (if (null (yic-ignore bn))
	  (setq go bf)
	(setq ptr (cdr ptr))
	)
      )
    (if go
	(switch-to-buffer go))))

(defun yic-prev-buffer ()
  "Vai para o buffer anterior na janela atual."
  (interactive)
  (yic-next (reverse (buffer-list))))

(defun yic-next-buffer ()
  "Vai para o proximo buffer, na janela atual."
  (interactive)
  (bury-buffer (current-buffer))
  (yic-next (buffer-list)))
;; fim das funcoes para troca de buffer
